const match = (input, originInput, ruleList) => {
  for (let rule of ruleList) {
    let match = null

    if (rule.match instanceof RegExp) {
      match = input.match(rule.match)
      match = match && match[0]
    }

    if (rule.match instanceof Function) {
      match = rule.match(input, originInput)
    }

    if (match) {
      return {
        match,
        type: rule.type,
        skip: rule.skip
      }
    }
  }
  return null
}

const tokenize = (input, ruleList) => {
  let offset = 0
  let matchToken = match(input.slice(offset), input, ruleList)
  const tokens = []

  // 初始token
  tokens.push({
    type: '^',
    match: '^',
    start: 0,
    end: 0
  })

  while (matchToken && offset < input.length) {
    if (!matchToken.skip) {
      tokens.push({
        type: matchToken.type,
        match: matchToken.match,
        start: offset,
        end: offset + matchToken.match.length,
      })
    }
    offset += matchToken.match.length
    matchToken = match(input.slice(offset), input, ruleList)
  }

  if (offset < input.length) {
    throw new Error(`Tokenize error: position ${offset}`)
  }

  // 结尾token
  tokens.push({
    type: '$',
    match: '$',
    start: input.length,
    end: input.length
  })

  return tokens
}

const tokenRuleList = [
  {
    type: 'command',
    match: /^@command_[a-zA-Z0-9]+/
  },
  {
    type: 'bracket',
    match: /^(\(|\))/
  },
  {
    type: 'space',
    match: /^\s+/,
    skip: true
  },
  {
    type: 'char',
    match: (input) => {
      return input[0]
    }
  }
]

const parse = (input) => {
  const tokens = tokenize(input, tokenRuleList)
  const tokensLength = tokens.length
  let index = 0
  let longestIndex = 0

  const castError = (msg = '') => {
    const newMsg = msg !== '' ? `parsing error in ${tokens[index].end}: ${msg}` : `parsing error: position ${tokens[longestIndex].start}`
    const error = new Error(newMsg)

    if (msg) {
      error.name = 'dead'
    }

    throw error
  }

  // 消费下一个token
  const read = () => {
    if (index >= tokensLength) {
      return null
    }
    const readToken = tokens[index]

    if (index > longestIndex) {
      longestIndex = index
    }
    index++
    return readToken
  }

  // 表达式式的token读取统一的收拢入口
  const readToken = (type, match) => {
    const token = read()
    if (token === null || token.type !== type || (match && token.match !== match)) {
      // token读取错误需要将未知回退
      if (token !== null) {
        index--
      }
      castError()
    }
    return token?.match
  }

  /**
   * 当存在Expr = Expr1 | Expr2 ｜ Expr3时，必须对Expr1出错时做一下处理
   * 1. 回溯到Expr2
   * 2. 重置index位置
   * 全部出错则表示Expr出错
   */
  const orRead = (...readings) => {
    let node = null
    const currentIndex = index

    for (let reading of readings) {
      try {
        node = reading()
        break
      } catch (e) {
        // 监测到打断回溯的错误，直接抛错
        if (e.name === 'dead') {
          throw e
        }
        index = currentIndex
      }
    }

    if (node === null) {
      castError()
    }

    return node
  }

  /**
   * allowEmpty = true, {[Expr]}, 表示Expr出现0次或多次
   * allowEmpty = false, {Expr}, 表示Expr出现多次
   */
  const multiRead = (reading, allowEmpty = true) => {
    const nodes = []
    let currentIndex = index;
    try {
      while (index < tokensLength) {
        const node = reading()
        nodes.push(node)
        currentIndex = index
      }
    } catch (e) {
      index = currentIndex
      if (e.name === 'dead') {
        throw e
      }
      if (!allowEmpty && nodes.length === 0) {
        castError()
      }
    }
    return nodes
  }

  const readGrammar = () => {
    readToken('^')
    const node = readStatements()
    readToken('$')
    return node
  }

  const readStatements = () => {
    return multiRead(readStatement)
  }

  const readStatement = () => {
    return orRead(
      readCommand,
      readLiteral
    )
  }

  const readCommand = () => {
    const node = {
      command: '',
      value: ''
    }
    const commandToken = readToken('command')
    node.command = commandToken?.split('@command_')[1]
    try {
      readToken('bracket', '(')
      node.value = readStatements()
      readToken('bracket', ')')
    } catch {
      castError('invalid command')
    }
    return node
  }

  const readLiteral = () => {
    const nodes = multiRead(() => readToken('char'), false)
    return nodes.join('')
  }

  return readGrammar()
}

const httpReg = new RegExp('^(?!mailto:)(?:(?:http|https|ftp)://|//)(?:\\S+(?::\\S*)?@)?(?:(?:(?:[1-9]\\d?|1\\d\\d|2[01]\\d|22[0-3])(?:\\.(?:1?\\d{1,2}|2[0-4]\\d|25[0-5])){2}(?:\\.(?:[0-9]\\d?|1\\d\\d|2[0-4]\\d|25[0-4]))|(?:(?:[a-z\\u00a1-\\uffff0-9]+-?)*[a-z\\u00a1-\\uffff0-9]+)(?:\\.(?:[a-z\\u00a1-\\uffff0-9]+-?)*[a-z\\u00a1-\\uffff0-9]+)*(?:\\.(?:[a-z\\u00a1-\\uffff]{2,})))|localhost)(?::\\d{2,5})?(?:(/|\\?|#)[^\\s]*)?$', 'i')

const commands = {
  encode: (str) => {
    return encodeURIComponent(removeInvalidParams(str))
  },
  h5path: (str) => {
    let isHttp = httpReg.test(str)
    const platform = getPlatform()

    if (isHttp) return `/web?url=${encodeURIComponent(removeInvalidParams(str))}`
    return `/web?path=${encodeURIComponent(removeInvalidParams(str))}`
  },
  shareLink: (str) => {
    const constReg = /\{\{([^\{\}]+)\}\}/
    const params = JSON.parse(str)
    Object.keys(params).forEach(key => {
      if (constReg.test(params[key])) delete params[key]
    })

    const platform = getPlatform()

    const data = {
      mpAppUserName: 'gh_f8c80781c128',
      trackSource: params.trackSource || 'web'
    }

    /* link & path都走小程序的h5 */
    if (params.link) data.path = `/pages/h5page/h5page?h5link=${encodeURIComponent(params.link)}`
    if (params.path) data.path = params.path

    /* android需要在encode一遍path */
    if (platform == "ANDROID") {
      data.path = encodeURIComponent(data.path)
    }

    if (params.poster) data.poster = params.poster
    if (params.title) data.title = params.title

    return `/shareInfo?data=${encodeURIComponent(JSON.stringify(data))}`
  }
}

const command = (path) => {
  const commandNodes = parse(path)
  const traverseNode = (node) => {
    if (typeof node === 'string') return node
    return commands[node.command](node.value.map(n => traverseNode(n)).join(''))
  }

  return commandNodes.map(node => traverseNode(node)).join('')
}

const removeInvalidParams = (path) => {
  const [_path, query] = path.split('?')
  const params = {}
  if (query) {
    query.split('&').forEach(param => {
      const [key, value] = param.split('=')
      if (/\{\{([^\{\}]+)\}\}/.test(value)) {
        return
      }
      params[key] = value || ''
    })
  }
  const nextQuery = Object.keys(params).map(key => `${key}=${params[key]}`).join('&')

  return nextQuery ? _path + '?' + nextQuery : _path
}

const pathReplace = (path, params) => {
  /* 常量替换 */
  const constReg = /\{\{([^\{\}]+)\}\}/g
  let replacedPath = path.replace(constReg, (matchStr, key) => {
    if (params && params[key]) {
      return params[key]
    }
    return matchStr
  })

  /* command执行 */
  replacedPath = command(replacedPath)

  return removeInvalidParams(replacedPath)
}